/* 
 * main.c --
 *
 *	This is a sample post-processor for the traces used in the
 *	Sprite paper entitled "Measurements of a Distributed File System".
 *
 * Copyright 1991 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

#ifndef lint
static char rcsid[] = "$Header: /local/src/cmds/sospexample/RCS/main.c,v 1.2 93/05/04 16:35:25 jhh Exp $ SPRITE (Berkeley)";
#endif /* not lint */

#include <sprite.h>
#include <stdio.h>
#include <stdlib.h>
#include <option.h>
#include <spriteTime.h>
#include <errno.h>
#include <sysStats.h>
#include <sosp.h>

char	*startTime = NULL;
char	*endTime = NULL;
char	*fileList = NULL;
char	*ignoreFile = NULL;
char	*directory = "/";
Boolean migrate = FALSE;
int	hoursToDo = -1;

Option optionArray[] = {
    {OPT_STRING, "s", (Address)&startTime, "Start time."},
    {OPT_STRING, "e", (Address)&endTime, "End time."},
    {OPT_STRING, "f", (Address) &fileList, "File of list of input files."},
    {OPT_STRING, "i", (Address) &ignoreFile, "List of files to ignore."},
    {OPT_STRING, "d", (Address) &directory,
	"Directory containing trace files."},
    {OPT_INT, "h", (Address) &hoursToDo,
	"Number of hours to process."},
};
int numOptions = Opt_Number(optionArray);
char 	*myname = NULL;

Time		time;
Time		firstTime;
Time		hourEnd;
Time		prevTime;
Time		nextTime;
int		bogusTime = 0;

extern long 	startSec;
extern long 	startUsec;
extern char	*ctime();

int	curSec;
int	curUsec;
int	nextSec;
int	nextUsec;

extern void bzero _ARGS_((char *ptr, int size));



/*
 *----------------------------------------------------------------------
 *
 * Init --
 *
 *	Initialize the package that reads trace records.  There are several
 * 	things that make this kind of gross.  First, the package expects
 *	an argc, argv string for some unknown reason, so we have to
 *	slap one together.  Second, the list of files to be processed
 *	is often too big for the shell so we have to allow them to
 *	be read from a file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
Init(argc, argv)
    int		argc;
    char	**argv;
{
    char 			**newargv;
    char 			**argvPtr;
    int				newargc;
    char			tmp[1024];
    char			tmp2[1024];
    char			*buf;
    int				num;
    FILE			*fd;
    int				i;
    ReturnStatus		status;

    /*
     * Allocate a new argv array (hope it's big enough) and fill it in.
     */
    newargv = (char **) malloc((2000) * (sizeof(char *)));
    argvPtr = newargv;
    newargc = 0;
    if (startTime != NULL) {
	*argvPtr++ = "-s";
	*argvPtr++ = startTime;
	newargc += 2;
    } 
    if (endTime != NULL) {
	*argvPtr++ = "-e";
	*argvPtr++ = endTime;
	newargc += 2;
    } 
    /*
     * Read the list of input files from a file.
     */
    if (fileList) {
	fd = fopen(fileList, "r");
	if (fd == NULL) {
	    fprintf(stderr, "Can't open %s: %s\n", fileList, 
		sys_errlist[errno]);
	    exit(1);
	}
	num = fscanf(fd, " %s", tmp);
	while(num != EOF) {
	    if (num != 1) {
		printf("Error reading list of files.\n");
		exit(1);
	    }
	    sprintf(tmp2, "%s/%s", directory, tmp); 
	    buf = malloc(strlen(tmp2) + 1);
	    strcpy(buf, tmp2);
	    *argvPtr++ = buf;
	    newargc++;
	    num = fscanf(fd, " %s", tmp);
	}
	fclose(fd);
    /* 
     * Or get the input files from argv.
     */
    } else {
	for (i = 1; i < argc; i++) {
	    *argvPtr++ = argv[i];
	    newargc++;
	}
    }
    status = initRead(myname, newargc, newargv, NULL);
    if (status != 0) {
	printf("%s: initRead failed 0x%x\n", myname, status);
	exit(1);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * LoadIgnoreFile --
 *
 *	Read the list of files to ignore from the input file and add
 *	them to the fid table.  Format of the input file is
 *	"host domain inode#".
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Stuff is loaded into fidInfoTable.
 *
 *----------------------------------------------------------------------
 */

void
LoadIgnoreFile(filename)
    char	*filename;
{
    FILE	*fdPtr;
    int		num;
    ID		id;
    int		line = 0;

    fdPtr = fopen(filename, "r");
    if (fdPtr == NULL) {
	fprintf(stderr, "Can't open %s : %s\n", filename, sys_errlist[errno]);
	exit(1);
    }
    num = fscanf(fdPtr, " %d %d %d ", &id.host, &id.domain, &id.inode);
    while(num != EOF) {
	line++;
	if (num == 3) {
	    /*
	     ************************************
	     *
	     * Do something here to keep track of files to be ignored.
	     * For example, you could put them in a hash table to check
	     * against:
	     *
	     *	entryPtr = Hash_CreateEntry(&skipFileIDToNum,
	     *		(Address) &id, NULL);
	     *
	     ************************************
	     */
	} else if (num != 0) {
	    fprintf(stderr, "Problem reading line %d of %s.\n", line, 
		filename);
	    exit(1);
	}
	num = fscanf(fdPtr, " %d %d %d ", &id.host, &id.domain, &id.inode);
    }
    fclose(fdPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
main(argc, argv)
int	argc;
char	**argv;
{
    int				i;
    int				status;
    RecordHeader		*recPtr;
    RecordHeader		*nextRecPtr;
    char			recBuf[1024];
    int				machine;
    int				recordCount;
    FILE			*fd;
    int				prevMachine;
    int				nextMachine;
    struct passwd		*passwdPtr;
    int				hours = 0;

    myname = argv[0];
    argc = Opt_Parse(argc, argv, optionArray, numOptions, 0);
    if (argc <= 0) {
	Opt_PrintUsage(myname, optionArray, numOptions);
	exit(1);
    }
    Init(argc, argv);
    if (ignoreFile) {
	LoadIgnoreFile(ignoreFile);
    }
    /*
     * Now start processing records.  We need to fetch two at a time
     * because sometimes time jumps forwards or backwards and we want
     * to ignore any records that have a weird time file.  Unfortunately,
     * getNextRecordMerge uses a static buffer, so we need to copy the
     * record to our own buffer before calling it the second time.
     * Also, the time fields in the trace record have to be incremented
     * by startSec and startUsec, which are global variables that are
     * modified during each call to getNextRecordMerge so we need to save
     * those two.  At the bottom of the loop we copy the next record
     * into the current record buffer and call getNextRecordMerge again.
     */
    machine = getNextRecordMerge((char **) &recPtr);
    if (machine < 0) {
	printf("%s: getNextRecordMerge failed.\n");
	exit(1);
    }
    bcopy((char *) recPtr, recBuf, recPtr->flags & TRACELOG_BYTEMASK);
    recPtr = (RecordHeader *) recBuf;
    curSec = startSec;
    curUsec = startUsec;
    firstTime.seconds = recPtr->time[0] + startSec;
    firstTime.microseconds = recPtr->time[1] + startUsec;
    printf("HOUR %d: %s", hours, ctime(&firstTime.seconds));
    Time_Add(firstTime, time_OneHour, &hourEnd);
    recordCount = 1;
    prevTime = firstTime;
    prevMachine = machine;
    nextMachine = getNextRecordMerge((char **) &nextRecPtr);
    if (machine < 0) {
	printf("%s: getNextRecordMerge failed.\n");
	exit(1);
    }
    while(machine > 0) {
	time.seconds = recPtr->time[0] + curSec;
	time.microseconds = recPtr->time[1] + curUsec;
	nextSec = startSec;
	nextUsec = startUsec;
	nextTime.seconds = nextRecPtr->time[0] + nextSec;
	nextTime.microseconds = nextRecPtr->time[1] + nextUsec;
	/*
	 * Sometimes time jumps forwards or backwards, but it doesn't
	 * happen very often so just ignore it when it does.
	 */
	if (Time_LT(time, prevTime)) {
	    bogusTime++;
	    goto nextRec;
	}
	if (Time_GT(time, nextTime)) {
	    bogusTime++;
	    goto nextRec;
	}
	/*
	 * Print something at the end of each hour so the user knows
	 * we're still doing something.
	 */
	if (Time_GE(time, hourEnd)) {
	    hours++;
	    printf("HOUR %d: %s", hours, ctime(&time.seconds));
	    Time_Add(hourEnd, time_OneHour, &hourEnd);
	}
	if (hours == hoursToDo) {
	    break;
	}
	/*
	 ***************************************************************
	 *
	 * REMOVE THIS IF YOU WANT:
	 * Print out contents of the trace record.
	 *
	 ***************************************************************
	 */
	dorec(recPtr);

	/*
	 * Fill in this switch statement with whatever you want to do.
	 */
	switch(recPtr->type) {
	    case SOSP_OPEN : {
		SospOpen	*openPtr;

		openPtr = (SospOpen *) recPtr;
		/*
		 **********************************************************
		 *
		 * Do stuff here to process an open event.  You may want to
		 * check that you only process open events for files that
		 * aren't swap files, directories, etc.  And you may want to
		 * ignore files that are dump files.  To do this, you will
		 * need to check for the following condition:
		 *
		 *  if (openPtr->type != FS_FILE) {
		 *	Ignore this one.
		 *  } else if (ignoreTraces) {
		 *      Ignore the event if its file ID was in ignoreFile.
		 *  } else if (!(openPtr->mode & FS_USER ||
		 *          (ignoreDumper && openPtr->realID == dumper)) {
		 *  }
		 *
		 * NOTE: the ignoreFile may contain the fileIDs of dump files,
		 * but for some of the traces, there was a special userID for
		 * the dumper, and this was recorded as part of the open event.
		 * The dumper's user ID is 28.
		 *
		 **********************************************************
		 */
		break;
	    }
	 }
	prevMachine = machine;
	prevTime = time;
nextRec:
	/*
	 * Copy the contents of the next record buffer into the current
	 * record buffer, then call getNextRecordMerge to get the next
	 * record.
	 */
	bcopy((char *) nextRecPtr, recBuf, 
		    nextRecPtr->flags & TRACELOG_BYTEMASK);
	machine = nextMachine;
	nextMachine = getNextRecordMerge((char **) &nextRecPtr);
	curSec = nextSec;
	curUsec = nextUsec;
	recordCount++;
    }
}
