#include <stdio.h>
#include <signal.h>
#include <sprite.h>
#include <sysStats.h>
#include "sosp.h"
#include <time.h>
#include <status.h>

extern long startSec;
extern long startUsec;


/*
 * Letters to keep track of trace versions.
 * Version 2: January 9
 */
#define VERSION1	'A'
#define VERSION2	'B'
#define VERSION3	'C'

#define VERSIONLETTER	VERSION2

extern int version;

char *opname(), *status(), *filetype();

dorec(tracePtr)
    Sys_TracelogRecord *tracePtr;
{
    int *data;
    int type;
    int i;
    long sec, usec;
    char *timeStr;
    int  len;
    data = (int *)&tracePtr->data;
    type = (tracePtr->recordLen&TRACELOG_TYPEMASK)>>16;
    len = (tracePtr->recordLen & TRACELOG_BYTEMASK);
    if (type != data[0]) {
	printf("Warning: type mismatch: %d vs %d\n", type, data[0]);
	type = data[0];
    }
    usec = tracePtr->time[1] + startUsec;
    sec = tracePtr->time[0] + startSec;
    if (usec>1000000) {
	sec++;
	usec -= 1000000;
    }
    timeStr = ctime((time_t *)&sec);
    timeStr[19] = '\0';
    printf("%s.%3.3d: ", timeStr+4, usec/1000);
    switch (type) {
	case SOSP_OPEN:
	    printf("OPEN: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x),\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    printf("\t\tstreamID: (%x,%x,%x,%x), effID: %d, realID %d,\n",
		    data[7], data[8], data[9], data[10], data[11], data[12]);
	    printf("\t\tmode %x, numNowReading: %d, numNowWriting: %d,\n",
		    data[13], data[14], data[15]);
	    printf("\t\tcreate: %d, fileSize: %d\n", data[16], data[17]);
	    if (len - 3 * sizeof(int) == 19 * sizeof(int)) {
		printf("\t\tmodify: %d\n", data[18]);
	    } else if (len - 3 * sizeof(int) == 20 * sizeof(int)) {
		printf("\t\tmodify: %d, type: %s\n", data[18],
			filetype(data[19]));
	    } else if (version == VERSION2) {
		printf("\t\tmodify: %d, type: %s, consist: %d\n", data[18],
			filetype(data[19]), data[20]);
	    }
	    break;
	case SOSP_DELETE:
	    printf("DELETE: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x),\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    if (len - 3 * sizeof(int) == 10 * sizeof(int)) {
		printf("\t\tmodify: %d, create: %d, size: %d\n", 
		    data[7], data[8], data[9]);
	    } else {
		printf("\t\tcreate: %d, fileSize: %d\n", data[7], data[8]);
	    }
	    break;
	case SOSP_CREATE:
	    printf("CREATE: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x)\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    break;
	case SOSP_MKLINK:
	    printf("MKLINK: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x)\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    break;
	case SOSP_SET_ATTR:
	    printf("SETATTR: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x)\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    break;
	case SOSP_GET_ATTR:
	    printf("GETATTR: HostID: %d, homeID: %d, fileID: (%x,%x,%x,%x)\n", 
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    break;
	case SOSP_LSEEK:
	    printf("LSEEK: streamID: (%x,%x,%x,%x), old offset: %d, new: %d\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    if (len - 3 * sizeof(int) == 8 * sizeof(int)) {
		printf("\t\trwFlags: 0x%x\n", data[7]);
	    }
	    break;
	case SOSP_CLOSE:
	    printf("CLOSE: streamID: (%x,%x,%x,%x), offset: %d, fileSize: %d\n",
		    data[1], data[2], data[3], data[4], data[5], data[6]);
	    if (len - 3 * sizeof(int) == 9 * sizeof(int)) {
		printf("\t\tflags: 0x%x, rwFlags: 0x%x\n", data[7], data[8]);
	    } else if (len - 3 * sizeof(int) == 10 * sizeof(int)) {
		printf("\t\tflags: 0x%x, rwFlags: 0x%x, refCount: 0x%x\n",
		    data[7], data[8], data[9]);
	    } else if (version==VERSION2) {
		printf("\t\tflags: 0x%x, rwFlags: 0x%x, refCount: 0x%x, consist: %d\n", 
		    data[7], data[8], data[9], data[10]);
		
	    } else {
		printf("\t\tflags: 0x%x\n", data[7]);
	    }

	    break;
	case SOSP_MIGRATE:
	    printf("MIGRATE: fromHostID: %d, toHostID: %d,\n",
		    data[1], data[2]);
	    printf("\t\tstreamID: (%x,%x,%x,%x), offset: %d\n",
		    data[3], data[4], data[5], data[6], data[7]);
	    break;
	case SOSP_TRUNCATE:
	    printf("TRUNCATE: streamID: (%x,%x,%x,%x),\n",
		data[1], data[2], data[3], data[4]);
	    printf("\t\toldLength: %d, newLength: %d\n", data[5], data[6]);
	    if (len - 3 * sizeof(int) == 9 * sizeof(int)) {
		printf("\t\tmodify: %d, create: %d\n", data[7], data[8]);
	    }
	    break;
	case SOSP_CONSIST_CHANGE:
	    printf("CONSIST_CHANGE: HostID: %d, fileID: (%x,%x,%x,%x),\n",
		    data[1], data[2], data[3], data[4], data[5]);
	    printf("\t\toperation: %d, forWrite?: %d\n", data[6], data[7]);
	    break;
	case SOSP_READ:
	    printf("READ: hostID: %d, streamID: (%x,%x,%x,%x),\n",
		    data[1], data[2], data[3], data[4], data[5]);
	    printf("\t\treadIt: %d, offset: %d, numbytes: %d\n",
		    data[6], data[7], data[8]);
	    break;
	case SOSP_LOOKUP:
	    printf("LOOKUP: hostID: %d, home: %d Status: %s, Op: %s",
		    data[1], data[2], status(data[7]), opname(data[9]&0xff));
	    if (version==VERSION2 && data[7]==SUCCESS) {
		printf(", Type: %s\n", filetype(data[9]>>8));
	    } else {
		printf("\n");
	    }
	    for (i=0;i<data[8];i++) {
		printf("\t\tFileID: (%x,%x,%x,%x)\n", data[10+i*4],
			data[11+i*4], data[12+i*4], data[13+i*4]);
	    }
	    printf("\t\tResultID: (%x,%x,%x,%x)\n", data[3], data[4],
		    data[5], data[6]);
	    break;
	case SOSP_CONSIST_ACTION:
	    printf("CONSIST_ACTION: causingHostID: %d, affectedHostID: %d,\n",
		    data[1], data[2]);
	    printf("\t\tfileID: (%x,%x,%x,%x), action: %x\n",
		    data[3], data[4], data[5], data[6], data[7]);
	    break;
	case SOSP_PREFIX:
	    printf("PREFIX: clientID: %d, rpcID: 0x%x\n", data[1], data[2]);
	    break;
	case LOST_TYPE:
	    printf("LOST RECORDS: %d\n", data[1]);
	    break;
	case SOSP_DELETE_DESC:
	    printf("DELETE_DESC: fileID: (%x,%x,%x,%x), modify: %d,\n",
		data[1], data[2], data[3], data[4], data[5]);
	    printf("\t\tcreate: %d, size: %d\n", data[6], data[7]);
	    break;

	default:
	    printf("Unknown type: %d\n", type);
    }
}

char buf[10];

char *opnames[] = {"Import", "Export", "Open", "GetAttrPath", "SetAttrPath",
     "MakeDevice", "MakeDir", "Remove", "RemoveDir", "Rename", "HardLink"};

char *opname(x)
int x;
{
    if (x>=0 && x<sizeof(opnames)/sizeof(char *)) {
	return opnames[x];
    } else {
	sprintf(buf,"%x",x);
	return buf;
    }
}

char *status(x)
int x;
{
     if (x==0) return "success";
     if (x==FS_FILE_NOT_FOUND) return "not found";
     if (x==FS_NO_ACCESS) return "no access";
     if (x==FS_LOOKUP_REDIRECT) return "lookup redirect";
     if (x==FS_IS_DIRECTORY) return "is directory";
     if (x==FS_NOT_DIRECTORY) return "not directory";
     if (x==FS_NOT_OWNER) return "not owner";
     if (x==FS_FILE_EXISTS) return "file exists";
     if (x==FS_DIR_NOT_EMPTY) return "dir not empty";
     sprintf(buf,"%x",x);
     return buf;
}

char *types[] = {"File", "Directory", "Symbolic link", "Remote link",
    "Device", "Remote device", "Local pipe", "Named pipe", "Pseudo dev",
    "Pseudo FS", "Extra file"};

char *filetype(x)
int x;
{
    if (x>=0 && x<sizeof(types)/sizeof(char *)) {
	return types[x];
    } else {
	sprintf(buf,"%x",x);
	return buf;
    }
}
